#![forbid(unsafe_code)]

use std::io::{stdout, Write, Read, stdin};

pub struct BrainFuckError<'a> {
    pub error: String,
    pub mem: &'a [u8]
}

impl BrainFuckError<'_> {
    pub fn new(error: String, mem: &[u8]) -> BrainFuckError {
        BrainFuckError { error, mem }
    }
}

/**
 * Check code for syntax errors
 */
pub fn lint_code(code: &&str) -> Result<(), Vec<String>> {

    let mut loop_stack: Vec<usize> = vec![];
    let mut pos = 0 as usize;

    let chars = code.chars().collect::<Vec<char>>();
    let size = code.len();

    let mut errors = vec![];

    loop {
        if pos == size {
            break
        }

        let char = chars[pos];
        pos += 1;

        if char == '[' {
            loop_stack.push(pos);
            continue
        }
        if char == ']' {
            if loop_stack.len() == 0 {
                errors.push(format!("ERR: Unmatched ']' ({})", pos));
                continue
            }
            loop_stack.pop();
        }
    }

    if loop_stack.len() != 0 {
        for lup in loop_stack.iter() {
            errors.push(format!("Unmatched '[' ({})", *lup));
        }
    }

    if errors.len() != 0 {
        return Err(errors)
    }

    Ok(())
}

pub fn eval_mem<'a>(bf_str: &&str, mem: &'a mut [u8], pointer: &mut usize) -> Result<(), BrainFuckError<'a>> {

    let mut pos = 0 as usize;
    let progsize = bf_str.len();
    let chars = bf_str.chars().collect::<Vec<char>>();
    
    let mut loop_stack: Vec<usize> = vec![];

    loop {
        if pos == progsize {
            break
        }

        let char = chars[pos];
        
        pos += 1;

        if char == '>' {
            if (*pointer) + 1 == 30000 {
                return Err(
                    BrainFuckError::new(format!("Attempt to increase pointer exceeding memory size ({})", pos), mem)
                );
            }
            *pointer += 1;
            continue
        }
        if char == '<' {
            if *pointer == 0 {
                return Err(
                    BrainFuckError::new(format!("Attempt to reduce pointer to less than 0 ({})", pos), mem)
                );
            }
            *pointer -= 1;
            continue
        }
        if char == '+' {
            if mem[*pointer] == 255 {

                #[cfg(feature = "safe-eval")]
                {
                    return Err(
                        BrainFuckError::new(format!("Attempt to overflow more than 255 ({})", pos), mem)
                    );
                }

                mem[*pointer] = 0;
                continue
            }
            mem[*pointer] += 1;
            continue
        }
        if char == '-' {
            if mem[*pointer] == 0 {

                #[cfg(feature = "safe-eval")]
                {
                    return Err(
                        BrainFuckError::new(format!("Attempt to overflow more than 255 ({})", pos), mem)
                    );
                }

                mem[*pointer] = 255;
                continue
            }
            mem[*pointer] -= 1;
            continue
        }
        if char == '.' {
            print!("{}", mem[*pointer] as char);
            stdout().flush().unwrap();
            continue
        }
        if char == ',' {
            let byte = stdin().bytes().next();
            if byte.is_none() {
                continue
            }
            let byte = byte.unwrap();
            if byte.is_err() {
                return Err(
                    BrainFuckError::new(byte.unwrap_err().to_string(), mem)
                );
            }
            mem[*pointer] = byte.unwrap();
            continue
        }
        if char == '[' {

            // if current is 0
            if mem[*pointer] == 0 {
                // skip the loop completely
                let mut p = pos.clone();
                loop {
                    if p == progsize {
                        return Err(
                            BrainFuckError::new(format!("Unmatched '[' ({})", pos), mem)
                        );
                    }

                    if chars[p] == ']' {
                        pos = p + 1;
                        break;
                    }

                    p += 1;
                }
                continue
            }

            loop_stack.push(pos);
            continue
        }
        if char == ']' {

            if loop_stack.len() == 0 {
                return Err(
                    BrainFuckError::new(format!("Unmatched ']' ({})", pos), mem)
                );
            }

            if mem[*pointer] == 0 {
                loop_stack.pop();
                continue
            } else {
                pos = loop_stack.last().unwrap().clone();
                continue
            }
        }
    }

    if loop_stack.len() != 0 {
        return Err(
            BrainFuckError::new(format!("Unmatched '[' ({})", loop_stack.last().unwrap()), mem)
        );
    }

    Ok(())
}

pub fn eval(bf_str: &&str) -> Result<[u8; 30000], String> {

    let mut memory: [u8; 30000] = core::array::from_fn(|_| 0);
    let mut pointer = 0 as usize;

    // first check the code

    let lint = lint_code(bf_str);
    if lint.is_err() {
        return Err(
            format!(
                "Code linting failed: \n\n{}",
                lint.unwrap_err().join("\n")
            )
        )
    }

    // then actually run it

    let e = eval_mem(bf_str, &mut memory, &mut pointer);
    if e.is_err() {
        let err = e.unwrap_err();
        return Err(err.error);
    }

    Ok(memory)
}

pub fn eval_no_lint<'a>(bf_str: &&str) -> Result<[u8; 30000], String> {
    let mut memory: [u8; 30000] = core::array::from_fn(|_| 0);
    let mut pointer = 0 as usize;

    let e = eval_mem(bf_str, &mut memory, &mut pointer);
    if e.is_err() {
        let err = e.unwrap_err();
        return Err(err.error);
    }

    Ok(memory)
}
